@@ -34,7 +34,6 @@
 #include "xorgVersion.h"
 
 #include "msm.h"
-#include "drm/drm_mode.h"
 #include "xf86Crtc.h"
 #include "xf86drmMode.h"
 #include "xf86DDC.h"
@@ -42,16 +41,12 @@
 #include "xf86RandR12.h"
 #include "X11/Xatom.h"
 
-#include "vbe.h"
 #include "dixstruct.h"
 #include "scrnintstr.h"
 #include "fb.h"
 #include "xf86cmap.h"
 #include "shadowfb.h"
 
-//#include "xf86xv.h"
-//#include <X11/extensions/Xv.h>
-
 #include "region.h"
 
 #include <X11/extensions/randr.h>
@@ -225,6 +220,106 @@ drmmode_crtc_dpms(xf86CrtcPtr drmmode_crtc, int mode)
 
 }
 
+static void
+drmmode_fbcon_copy(ScreenPtr pScreen)
+{
+	ScrnInfoPtr pScrn = xf86ScreenToScrn(pScreen);
+	MSMPtr pMsm = MSMPTR(pScrn);
+	ExaDriverPtr exa = pMsm->pExa;
+	drmmode_ptr drmmode = NULL;
+	xf86CrtcConfigPtr config = XF86_CRTC_CONFIG_PTR(pScrn);
+	drmModeFBPtr fb;
+	unsigned w = pScrn->virtualX, h = pScrn->virtualY, bpp = pScrn->bitsPerPixel;
+	uint32_t fbcon_id = 0;
+	struct fd_bo *fbcon_bo;
+	PixmapPtr fbcon_pix, scanout_pix;
+	void *ptr;
+	int i;
+
+	for (i = 0; i < config->num_crtc; i++) {
+		drmmode_crtc_private_ptr crtc = config->crtc[0]->driver_private;
+		if (crtc->mode_crtc->buffer_id) {
+			fbcon_id = crtc->mode_crtc->buffer_id;
+			drmmode = crtc->drmmode;
+		}
+	}
+
+	if (!fbcon_id)
+		goto fallback;
+
+	fb = drmModeGetFB(drmmode->fd, fbcon_id);
+	if (!fb) {
+		xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
+				"Failed to retrieve fbcon fb: id %d\n", fbcon_id);
+		goto fallback;
+	}
+
+	if (fb->depth != pScrn->depth || fb->width != w || fb->height != h) {
+		drmModeFreeFB(fb);
+		goto fallback;
+	}
+
+	fbcon_bo = fd_bo_from_handle(pMsm->dev, fb->handle,
+			fb->height * fb->pitch);
+	if (!fbcon_bo) {
+		xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
+				"Failed to retrieve fbcon buffer: handle=0x%08x\n",
+				fb->handle);
+		drmModeFreeFB(fb);
+		goto fallback;
+	}
+
+	fbcon_pix = drmmode_pixmap_wrap(pScreen, fb->width, fb->height,
+			fb->depth, fb->bpp, fb->pitch, fbcon_bo, NULL);
+	drmModeFreeFB(fb);
+	if (!fbcon_pix) {
+		xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
+				"Failed to create pixmap for fbcon contents\n");
+		goto fallback;
+	}
+
+	scanout_pix = drmmode_pixmap_wrap(pScreen, w, h, pScrn->depth,
+			bpp, pScrn->displayWidth * bpp / 8, pMsm->scanout, NULL);
+	if (!scanout_pix) {
+		xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
+				"Failed to init scanout pixmap for fbcon mirror\n");
+		pScreen->DestroyPixmap(fbcon_pix);
+		goto fallback;
+	}
+
+	/* we could use CopyArea() and get the normal sw fallbacks, but
+	 * then we need to do things in terms of drawables and regions.
+	 * So meh, we have XA, let's not worry too much about it.
+	 */
+	if (!exa->PrepareCopy(fbcon_pix, scanout_pix, 0, 0, GXcopy, ~0))
+		goto fallback;
+	exa->Copy(scanout_pix, 0, 0, 0, 0, w, h);
+	exa->DoneCopy(scanout_pix);
+
+	/* shouldn't really be needed, the PrepareAccess below will do this,
+	 * but for good measure:
+	 */
+	MSMFlushAccel(pScreen);
+
+	/* wait for completion before continuing, avoids seeing a momentary
+	 * flash of "corruption" on occasion
+	 */
+	exa->PrepareAccess(scanout_pix, EXA_PREPARE_SRC);
+	exa->FinishAccess(scanout_pix, EXA_PREPARE_SRC);
+
+	pScreen->DestroyPixmap(scanout_pix);
+	pScreen->DestroyPixmap(fbcon_pix);
+	pScreen->canDoBGNoneRoot = TRUE;
+
+	return;
+
+fallback:
+	ptr = fd_bo_map(pMsm->scanout);
+	if (!ptr)
+		return;
+	memset(ptr, 0x00, fd_bo_size(pMsm->scanout));
+}
+
 static Bool
 drmmode_set_mode_major(xf86CrtcPtr crtc, DisplayModePtr mode,
 		Rotation rotation, int x, int y)
@@ -241,9 +336,20 @@ drmmode_set_mode_major(xf86CrtcPtr crtc, DisplayModePtr mode,
 	int fb_id;
 	drmModeModeInfo kmode;
 
-	if (drmmode->fb_id == 0)
-		if (!drmmode_xf86crtc_resize(pScrn, mode->HDisplay, mode->VDisplay))
+	if (drmmode->fb_id == 0) {
+		int pitch = MSMAlignedStride(pScrn->virtualX,
+				pScrn->bitsPerPixel);
+		ret = drmModeAddFB(drmmode->fd, pScrn->virtualX, pScrn->virtualY,
+				pScrn->depth, pScrn->bitsPerPixel, pitch,
+				fd_bo_handle(pMsm->scanout), &drmmode->fb_id);
+		if (ret) {
+			xf86DrvMsg(crtc->scrn->scrnIndex, X_ERROR,
+					"Error adding FB for scanout: %s\n",
+					strerror(-ret));
 			return FALSE;
+		}
+		pScrn->displayWidth = pitch / (pScrn->bitsPerPixel >> 3);
+	}
 
 	if (!xf86CrtcRotate(crtc))
 		return FALSE;
@@ -324,20 +430,29 @@ convert_cursor(CARD32 *dst, CARD32 *src, int dw, int sw)
 	}
 }
 
-static void
-drmmode_load_cursor_argb (xf86CrtcPtr crtc, CARD32 *image)
+static Bool
+drmmode_load_cursor_argb_check(xf86CrtcPtr crtc, CARD32 *image)
 {
 	drmmode_crtc_private_ptr drmmode_crtc = crtc->driver_private;
 	struct fd_bo *cursor = drmmode_crtc->cursor;
 	drmmode_ptr drmmode = drmmode_crtc->drmmode;
 	void *ptr = fd_bo_map(cursor);
+	int ret = 0;
 
 	convert_cursor(ptr, image, 64, 64);
 
 	if (drmmode_crtc->cursor_visible) {
-		drmModeSetCursor(drmmode->fd, drmmode_crtc->mode_crtc->crtc_id,
+		ret = drmModeSetCursor(drmmode->fd, drmmode_crtc->mode_crtc->crtc_id,
 				fd_bo_handle(cursor), 64, 64);
 	}
+
+	return ret == 0;
+}
+
+static void
+drmmode_load_cursor_argb(xf86CrtcPtr crtc, CARD32 *image)
+{
+	drmmode_load_cursor_argb_check(crtc, image);
 }
 
 static void
@@ -468,6 +583,9 @@ static const xf86CrtcFuncsRec drmmode_crtc_funcs = {
 		.set_cursor_position = drmmode_set_cursor_position,
 		.show_cursor = drmmode_show_cursor,
 		.hide_cursor = drmmode_hide_cursor,
+#if XORG_VERSION_CURRENT >= XORG_VERSION_NUMERIC(1,15,99,902,0)
+		.load_cursor_argb_check = drmmode_load_cursor_argb_check,
+#endif
 		.load_cursor_argb = drmmode_load_cursor_argb,
 		.shadow_create = drmmode_crtc_shadow_create,
 		.shadow_allocate = drmmode_crtc_shadow_allocate,
@@ -619,7 +737,7 @@ drmmode_output_dpms(xf86OutputPtr output, int mode)
 
 	for (i = 0; i < koutput->count_props; i++) {
 		props = drmModeGetProperty(drmmode->fd, koutput->props[i]);
-		if (props && (props->flags && DRM_MODE_PROP_ENUM)) {
+		if (props && (props->flags & DRM_MODE_PROP_ENUM)) {
 			if (!strcmp(props->name, "DPMS")) {
 				mode_id = koutput->props[i];
 				drmModeFreeProperty(props);
@@ -891,6 +1009,8 @@ const char *output_names[] = {
 		"HDMI",
 		"TV",
 		"eDP",
+		"Virtual",
+		"DSI",
 };
 #define NUM_OUTPUT_NAMES (sizeof(output_names) / sizeof(output_names[0]))
 
@@ -980,7 +1100,7 @@ drmmode_xf86crtc_resize(ScrnInfoPtr pScrn, int width, int height)
 
 	if ((pScrn->virtualX == width) &&
 			(pScrn->virtualY == height) &&
-			pMsm->scanout)
+			pMsm->scanout && drmmode->fb_id)
 		return TRUE;
 
 	old_width = pScrn->virtualX;
@@ -1010,21 +1130,25 @@ drmmode_xf86crtc_resize(ScrnInfoPtr pScrn, int width, int height)
 			goto fail;
 	}
 
+	if (!old_fb_id) {
+		drmmode_fbcon_copy(screen);
+	} else {
+		memset(ptr, 0x00, fd_bo_size(pMsm->scanout));
+	}
+
 	/* NOTE do everything that could fail before this point,
 	 * otherwise you could end up w/ screen pixmap pointing
 	 * at the wrong scanout bo
 	 */
 	ppix = screen->GetScreenPixmap(screen);
 	if (ppix) {
-		msm_set_pixmap_bo(ppix, pMsm->scanout);
 		screen->ModifyPixmapHeader(ppix, width, height, -1, -1, pitch, ptr);
+		msm_set_pixmap_bo(ppix, pMsm->scanout);
 #if GET_ABI_MAJOR(ABI_VIDEODRV_VERSION) < 9
 		pScrn->pixmapPrivate.ptr = ppix->devPrivate.ptr;
 #endif
 	}
 
-	memset(ptr, 0x00, fd_bo_size(pMsm->scanout));
-
 	for (i = 0; i < xf86_config->num_crtc; i++) {
 		xf86CrtcPtr crtc = xf86_config->crtc[i];
 
@@ -1364,11 +1488,28 @@ drmmode_wait_for_event(ScrnInfoPtr pScrn)
 	drmHandleEvent(drmmode->fd, &drmmode->event_context);
 }
 
-void
+Bool
 drmmode_screen_init(ScreenPtr pScreen)
 {
 	ScrnInfoPtr pScrn = xf86ScreenToScrn(pScreen);
 	drmmode_ptr drmmode = drmmode_from_scrn(pScrn);
+	MSMPtr pMsm = MSMPTR(pScrn);
+	int pitch, size;
+
+	/* NOTE: we need an initial scanout buffer, in case no attached
+	 * display:
+	 */
+	pitch = MSMAlignedStride(pScrn->virtualX, pScrn->bitsPerPixel);
+	size = pitch * pScrn->virtualY;
+	DEBUG_MSG("initial scanout buffer: %dx%d@%d (size=%d, pitch=%d)",
+		pScrn->virtualX, pScrn->virtualY, pScrn->bitsPerPixel,
+		size, pitch);
+	pMsm->scanout = fd_bo_new(pMsm->dev, size,
+			DRM_FREEDRENO_GEM_TYPE_KMEM);
+	if (!pMsm->scanout) {
+		ERROR_MSG("Error allocating scanout buffer");
+		return FALSE;
+	}
 
 	drmmode_uevent_init(pScrn);
 
@@ -1381,6 +1522,8 @@ drmmode_screen_init(ScreenPtr pScreen)
 	/* Register a wakeup handler to get informed on DRM events */
 	RegisterBlockAndWakeupHandlers((BlockHandlerProcPtr)NoopDDA,
 			drmmode_wakeup_handler, pScrn);
+
+	return TRUE;
 }
 
 void
@@ -1397,6 +1540,7 @@ drmmode_screen_fini(ScreenPtr pScreen)
 			drmmode_wakeup_handler, pScrn);
 	RemoveGeneralSocket(drmmode->fd);
 
+	drmmode_remove_fb(pScrn);
 	fd_bo_del(pMsm->scanout);
 	pMsm->scanout = NULL;
 }
